#ifndef SLIST_H
#define SLIST_H

#include <memory>
#include <cassert>

template<class T, class Allocator = std::allocator<T> >
class slist {
	struct Item;
	typedef typename Allocator::rebind<Item>::other Item_Allocator;
public:
	// Anforderungen an alle Container
	typedef typename Allocator::value_type value_type;
	typedef typename Allocator::reference reference;
	typedef typename Allocator::const_reference const_reference;
	class iterator;
	class const_iterator;
	typedef typename Allocator::difference_type difference_type;
	typedef typename Allocator::size_type size_type;
	typedef typename Allocator::pointer pointer;
	typedef typename Allocator::const_pointer const_pointer;
	typedef Allocator allocator_type;
	explicit slist(const Allocator& alloc = Allocator());
	slist(const slist& x);
	~slist();
	slist& operator=(const slist& x);
	iterator begin() { return iterator(first); }
	const_iterator begin() const { return const_iterator(first); }
	iterator end() { return iterator(0); }
	const_iterator end() const { return const_iterator(0); }
	size_type size() const;
	size_type max_size() const;
	bool empty() const { return first == 0; }
	bool operator==(const slist&) const;
	bool operator!=(const slist& x) const { return !(*this == x); }
	bool operator<(const slist&) const;
	bool operator>(const slist& x) const { return x < *this; }
	bool operator<=(const slist& x) const { return !(x < *this); }
	bool operator>=(const slist& x) const { return !(*this < x ); }
	void swap(slist&);
	allocator_type get_allocator() const { return value_allocator; }
	// Anforderungen an sequenzielle Container werden teilweise erfllt
	explicit slist(size_type n, const T& value = T(),
		const Allocator& alloc = Allocator());
	template<class InputIterator>
		slist(InputIterator first, InputIterator last,
			const Allocator& alloc = Allocator());
	void clear();
	// Ersatz fr insert und erase
	template<class InputIterator>
		void insert_after(iterator position, InputIterator first, InputIterator last);
	iterator insert_after(iterator position, const T& x = T());
	void insert_after(iterator position, size_type n, const T& x);
	void erase_after(iterator position);
	void erase_after(iterator position, iterator last);
	// optionale Anforderungen fr sequenzielle Container
	reference front() { return first->data; }
	const_reference front() const { return first->data; }
	void push_front(const T& x);
	void pop_front();
	// weitere typische Funktionen fr sequenzielle Container
	void assign(size_type n, const T& u = T());
	template<class InputIterator>
		void assign(InputIterator first, InputIterator last);
	void resize(size_type sz, T c = T());
	// spezielle Listenfunktionen
	// Ersatz fr splice
	void splice_after(iterator position, slist& x);
	void splice_after(iterator position, slist& x, iterator i);
	void splice_after(iterator position, slist& x, iterator first, iterator last);
	void remove(const T& value);
	template<class Predicate> void remove_if(Predicate pred);
	void unique();
	template<class BinaryPredicate> void unique(BinaryPredicate binary_pred);
	void merge(slist& x);
	template<class Compare> void merge(slist& x, Compare comp);
	void reverse();
	void sort();
	template<class Compare> void sort(Compare comp);
	// Definition der Iteratorklassen
	class iterator : public std::iterator<std::forward_iterator_tag, value_type> {
	public:
		iterator() : current(0) { }
		bool operator==(const iterator& x) const { return current == x.current; }
		bool operator!=(const iterator& x) const
			{ return current != x.current; }
		const_reference operator*() const { return current->data; }
		reference operator*() { return current->data; }
		pointer operator->() { return &current->data; }
		const_pointer operator->() const { return &current->data; }
		iterator& operator++() { current = current->next; return *this; }
		iterator operator++(int) { iterator tmp(*this); ++*this; return tmp; }
	private:
		friend class slist<T, Allocator>;
		friend class const_iterator;
		explicit iterator(typename Item_Allocator::pointer p) : current(p) { }
		typename Item_Allocator::pointer current;
	};
	class const_iterator : public std::iterator<std::forward_iterator_tag, value_type> {
	public:
		const_iterator() : current(0) { }
		const_iterator(const iterator& i) : current(i.current) { }
		bool operator==(const const_iterator& x) const
			{ return current == x.current; }
		bool operator!=(const const_iterator& x) const
			{ return current != x.current; }
		const_reference operator*() const { return current->data; }
		const_pointer  operator->() const { return &current->data; }
		const_iterator& operator++() { current = current->next; return *this; }
		const_iterator operator++(int)
			{ const_iterator tmp(*this); ++*this; return tmp; }
	private:
		friend class slist<T, Allocator>;
		explicit const_iterator(typename Item_Allocator::const_pointer p)
			: current(p) { }
		typename Item_Allocator::const_pointer current;
	};
private:
	typename Item_Allocator::pointer new_item(const T& x,
		typename Item_Allocator::pointer p);
	void delete_item(typename Item_Allocator::pointer);
	struct Item {
		explicit Item(const_reference d, typename Item_Allocator::pointer n = 0)
			: data(d), next(n) { }
		value_type data;
		typename Item_Allocator::pointer next;
	};
	typename Item_Allocator::pointer first;
	Allocator value_allocator;
	Item_Allocator item_allocator;
};

template<class T, class Allocator>
typename slist<T, Allocator>::Item_Allocator::pointer
slist<T, Allocator>::new_item(const T& x, typename Item_Allocator::pointer p) {
	typename Item_Allocator::pointer tmp = item_allocator.allocate(1);
	item_allocator.construct(tmp, Item(x, p));
	return tmp;
}

template<class T, class Allocator>
void slist<T, Allocator>::delete_item(typename Item_Allocator::pointer p) {
	item_allocator.destroy(p);
	item_allocator.deallocate(p, 1);
}

template<class T, class Allocator> inline
void swap(slist<T, Allocator>& x, slist<T, Allocator>& y) { x.swap(y); }

template<class T, class Allocator> inline
slist<T, Allocator>::slist(const Allocator& alloc)
	: first(0), value_allocator(alloc), item_allocator(alloc) { }

template<class T, class Allocator> inline
slist<T, Allocator>::~slist() { clear(); }

template<class T, class Allocator> inline
slist<T, Allocator>::size_type slist<T, Allocator>::max_size() const
	{ return value_allocator.max_size(); }

template<class T, class Allocator> inline
void slist<T, Allocator>::swap(slist& x) { std::swap(first, x.first); }

template<class T, class Allocator> inline
void slist<T, Allocator>::push_front(const T& x)
	{ first = new_item(x, first); }

template<class T, class Allocator> inline
void slist<T, Allocator>::pop_front() {
	Item_Allocator::pointer tmp = first;
	first = first->next;
	delete_item(tmp);
}

template<class T, class Allocator>
void slist<T, Allocator>::clear() { while (first != 0) pop_front(); }

template<class T, class Allocator>
slist<T, Allocator>::size_type slist<T, Allocator>::size() const {
	size_type n = 0;
	for (Item_Allocator::const_pointer p = first; p != 0; p = p->next) ++n;
	return n;
}

template<class T, class Allocator>
slist<T, Allocator>::slist(size_type n, const T& value, const Allocator& alloc)
	: first(0), value_allocator(alloc), item_allocator(alloc) {
	while (n-- > 0)
		push_front(value);
}

template<class T, class Allocator>
slist<T, Allocator>::slist(const slist& x)
	: first(0), value_allocator(x.value_allocator),
	item_allocator(x.item_allocator) {
	if (!x.empty()) {
		push_front(*x.begin());
		insert_after(begin(), ++x.begin(), x.end());
	}
	assert(*this == x);
}

template<class T, class Allocator>
template<class InputIterator>
slist<T, Allocator>::slist(InputIterator f, InputIterator l,
	const Allocator& alloc) : first(0), value_allocator(alloc), item_allocator(alloc) {
	if (f != l) {
		push_front(*f);
		insert_after(begin(), ++f, l);
	}
}

template<class T, class Allocator>
slist<T, Allocator>& slist<T, Allocator>::operator=(const slist& x) {
	if (this == &x) return *this;
	Item_Allocator::const_pointer p = x.first;
	Item_Allocator::pointer q = first, r = 0;
	while (p != 0 && q != 0) {
		q->data = p->data;
		r = q;
		q = q->next;
		p = p->next;
	}
	if (q != 0)  // alte Liste lnger als die neue
		erase_after(iterator(r), end());
	else if (p != 0) {  // alte Liste krzer als die neue
		if (r == 0) { // alte Liste ist leer
			push_front(p->data);
			r = first;
			p = p->next;
		}
		insert_after(iterator(r), const_iterator(p), x.end());
	}
	assert(*this == x);
	return *this;
}

template<class T, class Allocator>
template<class InputIterator>
void slist<T, Allocator>::assign(InputIterator f, InputIterator l) {
	Item_Allocator::pointer q = first, r = 0;
	while (f != l && q != 0) {
		q->data = *f++;
		r = q;
		q = q->next;
	}
	if (q != 0)  // alte Liste lnger als die neue
		erase_after(iterator(r), end());
	else if (f != l) {  // alte Liste krzer als die neue
		if (r == 0) { // alte Liste ist leer
			push_front(*f++);
			r = first;
		}
		insert_after(iterator(r), f, l);
	}
}

template<class T, class Allocator>
void slist<T, Allocator>::assign(size_type n, const T& u) {
	// in die vorhandenen Listenelemente den Wert u kopieren
	Item_Allocator::pointer p = first,
		q = 0;  // zum Merken des letzten Elements mit dem Wert u
	while (p != 0 && n-- > 0) {
		p->data = u;
		q = p;
		p = p->next;
	}
	if (p != 0) {  // ggf. berzhlige Listenelemente nach q also ab p lschen
		if (q == 0)  // gleichbedeutend mit n == 0 und p == first
			first = 0;
		else
			erase_after(iterator(q), end());
	} else             // ggf. fehlende Listenelemente vorne einfgen
		while (n-- > 0)  // nutzt aus, da alle Elemente denselben Wert erhalten
			push_front(u);
}

template<class T, class Allocator>
void slist<T, Allocator>::resize(size_type n, T c) {
	Item_Allocator::pointer p = first, q = 0;  // zum Merken des letzten Elements
	while (p != 0 && n-- > 0) {
		q = p;
		p = p->next;
	}
	if (p != 0) {  // ggf. berzhlige Listenelemente nach q also ab p lschen
		if (q == 0)  // gleichbedeutend mit n == 0 und p == first
			first = 0;
		else
			erase_after(iterator(q), end());
	} else {         // ggf. fehlende Listenelemente anhngen
		if (q == 0) {  // alte Liste ist leer
			push_front(c);
			q = first;
			--n;
		}
		iterator i(q);
		while (n-- > 0)
			i = insert_after(i, c);
	}
}

template<class T, class Allocator>
void slist<T, Allocator>::remove(const T& value) {
	// Solange das jeweils erste Element lschen, wie der Wert gleich value ist.
	while (!empty() && front() == value)
		pop_front();
	// Jetzt ist die Liste entweder leer oder das erste Element ist ungleich
	// value und bleibt deshalb in der Liste.
	if (!empty()) {  // Liste nicht leer
		// Ausgehend von einem Listenelement wird jeweils das nchste untersucht
		// und ggf. gelscht, so wird kein zustzlicher Zeiger bentigt.
		Item_Allocator::pointer p = first;
		while (p->next != 0)
			if (p->next->data == value)
				erase_after(iterator(p));
			else
				p = p->next;
	}
}

template<class T, class Allocator>
void slist<T, Allocator>::unique() {
	// Das erste Listenelement bleibt in jedem Fall erhalten.
	Item_Allocator::pointer p = first;
	while (p != 0 && p->next != 0)
		if (p->data == p->next->data)
			erase_after(iterator(p));
		else
			p = p->next;
}

template<class T, class Allocator>
void slist<T, Allocator>::reverse() {
	Item_Allocator::pointer p0 = 0, p1 = first;
	while (p1 != 0) {
		Item_Allocator::pointer p2 = p1->next;
		p1->next = p0;
		p0 = p1;
		p1 = p2;
	}
	first = p0;
}

template<class T, class Allocator> inline
void slist<T, Allocator>::erase_after(iterator position) {
	// entfernt das Element hinter position
	Item_Allocator::pointer tmp = position.current->next;
	position.current->next = tmp->next;
	delete_item(tmp);
}

template<class T, class Allocator>
void slist<T, Allocator>::erase_after(iterator position, iterator l) {
	// entfernt die Elemente des Bereichs (position, l)
	while (position.current->next != l.current)
		erase_after(position);
}

template<class T, class Allocator> inline
slist<T, Allocator>::iterator
slist<T, Allocator>::insert_after(iterator position, const T& x) {
	// fgt ein neues Element mit dem Wert x hinter position ein
	position.current->next = new_item(x, position.current->next);
	return iterator(position.current->next);
}

template<class T, class Allocator>
void
slist<T, Allocator>::insert_after(iterator position, size_type n, const T& x) {
	// fgt hinter position n Elemente mit dem Wert x ein
	while (n-- > 0)
		insert_after(position, x);
}

template<class T, class Allocator>
template<class InputIterator>
void slist<T, Allocator>::insert_after(iterator pos, InputIterator f,
	InputIterator l) {
	// fgt die Elemente des Bereichs [f, l) hinter pos ein
	while (f != l)
		pos = insert_after(pos, *f++);
}

template<class T, class Allocator>
void slist<T, Allocator>::splice_after(iterator position, slist& x) {
	// alle Elemente von x aus x entfernen und hinter position einfgen
	assert(position.current != 0);
	if (!x.empty()) {  // wenn x keine Elemente enthlt, gibt es nichts zu tun
		Item_Allocator::pointer p = x.first;  // letztes Element von x ermitteln
		while (p->next != 0) p = p->next;
		p->next = position.current->next;
		position.current->next = x.first;
		x.first = 0;  // x besitzt keine Elemente mehr
	}
}

template<class T, class Allocator>
void
slist<T, Allocator>::splice_after(iterator position, slist& /*x*/, iterator i) {
	// das Element aus x, das nach i kommt, aus x entfernen
	// und hinter position einfgen
	assert(position.current != 0 && i.current->next != 0);
	Item_Allocator::pointer tmp = i.current->next->next;
	i.current->next->next = position.current->next;
	position.current->next = i.current->next;
	i.current->next = tmp;
}

template<class T, class Allocator>
void slist<T, Allocator>::splice_after(iterator position, slist& /*x*/,
	iterator f, iterator l) {
	// die Elemente des Bereichs (f, l] aus x entfernen
	// und hinter position einfgen
	Item_Allocator::pointer tmp = l.current->next;
	l.current->next = position.current->next;
	position.current->next = f.current->next;
	f.current->next = tmp;
}

template<class T, class Allocator>
bool slist<T, Allocator>::operator==(const slist<T, Allocator>& x) const {
	// gleich, wenn gleiche Anzahl von Elementen mit gleichen Werten
	if (this == &x) return true;
	Item_Allocator::const_pointer q = first, p = x.first;
	while (q != 0 && p != 0 && p->data == q->data) {
		p = p->next;
		q = q->next;
	}
	return q == 0 && p == 0;
}

template<class T, class Allocator>
bool slist<T, Allocator>::operator<(const slist<T, Allocator>& x) const {
	if (this == &x) return false;
	return lexicographical_compare(begin(), end(), x.begin(), x.end());
}

template<class T, class Allocator>
void slist<T, Allocator>::sort() {  // O(n*n)
	slist s;
	for (iterator i = begin(); i != end(); ++i) {
		iterator j = s.begin(), k = s.end();
		while (j != s.end() && *j < *i)
			k = j++;
		if (k == s.end())
			s.push_front(*i);
		else
			s.insert_after(k, *i);
	}
	swap(s);
}

template<class T, class Allocator>
void slist<T, Allocator>::merge(slist& x) {
	if (empty()) {
		swap(x);
		return;
	}
	Item_Allocator::pointer p = first;
	while (x.first != 0) {
		// in *this solange laufen, bis  q < x.first < p gilt
		Item_Allocator::pointer q = 0;
		while (p != 0 && p->data < x.first->data) {
			q = p;
			p = p->next;
		}
		if (p != 0) {
			// in x solange laufen, bis r < p < r->next gilt
			Item_Allocator::pointer r = x.first;
			while (r->next != 0 && r->next->data < p->data)
				r = r->next;
			// q < x.first < ... < r < p < r->next
			if (q != 0)
				q->next = x.first;
			else  // das erste Element von x ist das kleinste
				first = x.first;
			x.first = r->next;
			r->next = p;
		} else {  // am Ende von *this
			q->next = x.first;
			x.first = 0;
		}
	}
}

template<class T, class Allocator>
template<class Predicate>
void slist<T, Allocator>::remove_if(Predicate pred) {
	// Solange das jeweils erste Element lschen, wie der Wert gleich value ist.
	while (!empty() && pred(front()))
		pop_front();
	// Jetzt ist die Liste entweder leer oder das erste Element ist ungleich
	// value und bleibt deshalb in der Liste.
	if (!empty()) {
		// Ausgehend von einem Listenelement wird jeweils das nchste untersucht
		// und ggf. gelscht, so wird kein zustzlicher Zeiger bentigt.
		Item_Allocator::pointer p = first;
		while (p->next != 0)
			if (pred(p->next->data))
				erase_after(iterator(p));
			else
				p = p->next;
	}
}

template<class T, class Allocator>
template<class BinaryPredicate>
void slist<T, Allocator>::unique(BinaryPredicate binary_pred) {
	// Das erste Listenelement bleibt in jedem Fall erhalten.
	Item_Allocator::pointer p = first;
	while (p != 0 && p->next != 0)
		if (binary_pred(p->data, p->next->data))
			erase_after(iterator(p));
		else
			p = p->next;
}

template<class T, class Allocator>
template<class Compare>
void slist<T, Allocator>::merge(slist& x, Compare comp) {
	if (empty()) {
		swap(x);
		return;
	}
	Item_Allocator::pointer p = first;
	while (x.first != 0) {
		// In *this solange laufen, bis  q < x.first < p gilt.
		Item_Allocator::pointer q = 0;
		while (p != 0 && comp(p->data, x.first->data)) {
			q = p;
			p = p->next;
		}
		if (p != 0) {
			// In x solange laufen bis r < p < r->next gilt.
			Item_Allocator::pointer r = x.first;
			while (r->next != 0 && comp(r->next->data, p->data))
				r = r->next;
			// q < x.first < ... < r < p < r->next
			if (q != 0)
				q->next = x.first;
			else  // Das erste Element von x ist das kleinste.
				first = x.first;
			x.first = r->next;
			r->next = p;
		} else {  // am Ende von *this
			q->next = x.first;
			x.first = 0;
		}
	}
}

template<class T, class Allocator>
template<class Compare>
void slist<T, Allocator>::sort(Compare comp) {
	slist s;
	for (iterator i = begin(); i != end(); ++i) {
		iterator j = s.begin(), k = s.end();
		while (j != s.end() && comp(*j, *i))
			k = j++;
		if (k == s.end())
			s.push_front(*i);
		else
			s.insert_after(k, *i);
	}
	swap(s);
}

#endif
